/*
 * $Id: dot1dTp.c,v 1.1.2.9 2006/11/06 04:17:20 bobyu Exp $
 */

/* This should always be included first before anything else */
#include <config.h>

/* minimal include directives */
#include "mibincl.h"
#include "mibgroup/util_funcs.h"

#include "mibgroup/mibII/dot1dTp.h"
#include <ucd-snmp/mibgroup/mibII/bridge_mib.h>

/* Parameters to the callback function */
static bridge_parameters_t *dot1dtp_parms = NULL;

/* Path to this MIB WRT global MIB tree */
oid dot1dTp_variables_oid[] = { 1,3,6,1,2,1,17,4 };

/* index of entry under consideration */
static int tp_entry_num;

/* Variables that this MIB Supports */
struct variable4 dot1dTp_variables[] = {
/*  magic number        , variable type , ro/rw , callback fn  , L, oidsuffix */
#define DOT1DTPLEARNEDENTRYDISCARDS     1
    {DOT1DTPLEARNEDENTRYDISCARDS,  ASN_COUNTER,  RONLY ,  var_dot1dTp, 1,  { 1 }},
#define DOT1DTPAGINGTIME        2
    {DOT1DTPAGINGTIME,  ASN_INTEGER,  RWRITE,  var_dot1dTp, 1,  { 2 }},
#define DOT1DTPFDBADDRESS       3
    {DOT1DTPFDBADDRESS,  ASN_OCTET_STR,  RONLY,   var_dot1dTpFdbTable, 3,  { 3, 1, 1 }},
#define DOT1DTPFDBPORT      4
    {DOT1DTPFDBPORT,  ASN_INTEGER,  RONLY,   var_dot1dTpFdbTable, 3,  { 3, 1, 2 }},
#define DOT1DTPFDBSTATUS        5
    {DOT1DTPFDBSTATUS,  ASN_INTEGER,  RONLY,   var_dot1dTpFdbTable, 3,  { 3, 1, 3 }},
#define DOT1DTPPORT     6
    {DOT1DTPPORT,  ASN_INTEGER,  RONLY,   var_dot1dTpPortTable, 3,  { 4, 1, 1 }},
#define DOT1DTPPORTMAXINFO      7
    {DOT1DTPPORTMAXINFO,  ASN_INTEGER,  RONLY,   var_dot1dTpPortTable, 3,  { 4, 1, 2 }},
#define DOT1DTPPORTINFRAMES     8
    {DOT1DTPPORTINFRAMES,  ASN_COUNTER,  RONLY,   var_dot1dTpPortTable, 3,  { 4, 1, 3 }},
#define DOT1DTPPORTOUTFRAMES        9
    {DOT1DTPPORTOUTFRAMES,  ASN_COUNTER,  RONLY,   var_dot1dTpPortTable, 3,  { 4, 1, 4 }},
#define DOT1DTPPORTINDISCARDS       10
    {DOT1DTPPORTINDISCARDS,  ASN_COUNTER,  RONLY,   var_dot1dTpPortTable, 3,  { 4, 1, 5 }},
};

/** Initializes the dot1dTp module */
void
init_dot1dTp(void)
{
    dot1dtp_parms = (bridge_parameters_t *) \
                sal_alloc(sizeof(bridge_parameters_t), MIBII_DOT1DTP_MIB);

    if (dot1dtp_parms != NULL) {
        /* register ourselves with the agent to handle our mib tree */
        REGISTER_MIB(MIBII_DOT1DTP_MIB, dot1dTp_variables, variable4,  
               dot1dTp_variables_oid);
        
        dot1dtp_parms->group = BRIDGE_DOT1DTP;
    }
}


/*
 * var_dot1dTp():
 */
unsigned char *
var_dot1dTp(struct variable *vp, 
                oid     *name, 
                size_t  *length, 
                int     exact, 
                size_t  *var_len, 
                WriteMethod **write_method)
{
    /* variables we may use later */
    static long long_ret;

    if (header_generic(vp,name,length,exact,var_len,write_method)
                                  == MATCH_FAILED ) {
        AGENT_EXIT;
        return NULL;
    }

    switch(vp->magic) {
        case DOT1DTPLEARNEDENTRYDISCARDS:
            dot1dtp_parms->group_obj = DOT1D_TPLEARNEDENTRYDISCARDS;
            break;

        case DOT1DTPAGINGTIME:
            *write_method = write_dot1dTpAgingTime;
            dot1dtp_parms->group_obj = DOT1D_TPAGINGTIME;
            break;

        default:
            AGENT_EXIT;
            return NULL;
    }

    /* Callback */
    dot1dtp_parms->callback_rw = CALLBACK_ACT_READ;
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                        SNMPD_CALLBACK_BRIDGE,
                        dot1dtp_parms);
    /* Give the value back */
    long_ret = (long)dot1dtp_parms->value;
    return (unsigned char *)&long_ret;
}

static int
header_dot1dTpFdbTable(struct variable *vp,
           oid *name,
           size_t *length,
           int exact,
           size_t *var_len,
           WriteMethod **write_method)
{
#define DOT1DTPFDB_NAME_LENGTH  11
#define DOT1DTPFDB_TOTAL_OID_LENGTH 17

    oid newname[MAX_OID_LEN];
    oid lowest[MAX_OID_LEN];
    int i, result, found = FALSE;
    int num_mac_addr;
    unsigned char mac[6];
    int j = 0;

    memcpy( (char *)newname,(char *)vp->name, (int)vp->namelen * sizeof(oid));
    snmp_mib_setInterfaceCount();

    /* Set the entry number to an invalid index at the start */
    tp_entry_num = -1;

    /* Get count of number of L2 entries */
    dot1dtp_parms->group = BRIDGE_DOT1DSTATIC;
    
    /* dump l2 entries first */
    dot1dtp_parms->group_obj = DOT1D_STATICADDRESSDUMP;
    dot1dtp_parms->callback_rw = CALLBACK_ACT_READ;
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                        SNMPD_CALLBACK_BRIDGE,
                        dot1dtp_parms);

    dot1dtp_parms->group_obj = DOT1D_STATICADDRESSCOUNT;
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                        SNMPD_CALLBACK_BRIDGE,
                        dot1dtp_parms);
    num_mac_addr = dot1dtp_parms->value;
    AGENT_DBG(("No. of fdb entries are %d", num_mac_addr));

    /* Now, get the entry by index and validate request */
    for (i = 0; i < num_mac_addr; i++) {
        /* get the mac address by index */
        dot1dtp_parms->index = i;
        dot1dtp_parms->group_obj = DOT1D_STATICADDRESS;
        dot1dtp_parms->callback_rw = CALLBACK_ACT_READ;
        snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                            SNMPD_CALLBACK_BRIDGE,
                            dot1dtp_parms);
        memcpy (mac, dot1dtp_parms->buf, MAC_ADDRESS_LEN);

        if ( (mac[0] & 0x1)) {
            DEBUGMSGTL(("dot1dTp", "Not a unicast entry\n"));
            continue;
        }

        for (j = 0; j < MAC_ADDRESS_LEN; j++)  { /* 6 octets for MAC Address */
            AGENT_DBG(("Mac address at %d is %x", j, mac[j]));
            newname[DOT1DTPFDB_NAME_LENGTH+j] = mac[j];
        }

        /* if it's a SET/GET request, exact is true. So return only if there
         * is an exact match or snmp_oid_compare returns 0
         */
        if (exact) {
            if (snmp_oid_compare(name, *length, newname, 
                        DOT1DTPFDB_TOTAL_OID_LENGTH) == 0) {
                memcpy(lowest, newname, 
                                DOT1DTPFDB_TOTAL_OID_LENGTH * sizeof(oid));
                found = TRUE;
                tp_entry_num = i;
                break;
            }
        } else {
            if (snmp_oid_compare(newname, DOT1DTPFDB_TOTAL_OID_LENGTH, 
                        name, *length) > 0 && (found == FALSE ||
                snmp_oid_compare(newname, DOT1DTPFDB_TOTAL_OID_LENGTH, lowest, 
                    DOT1DTPFDB_TOTAL_OID_LENGTH) < 0)) {
                /* Determine if get or get-next */
                memcpy((char *) lowest, (char *) newname,
                       DOT1DTPFDB_TOTAL_OID_LENGTH * sizeof(oid));
                found = TRUE;
                tp_entry_num = i;
            }
        }
    }
    dot1dtp_parms->group = BRIDGE_DOT1DTP;

    if (!found) {
        DEBUGMSGTL(("dot1dTp", "... index out of range\n"));
        return MATCH_FAILED;
    }
 
    /* Let us set the write method */
    *write_method = 0;
    memcpy( (char *)name,(char *)lowest, 
            DOT1DTPFDB_TOTAL_OID_LENGTH * sizeof(oid));

    *length = DOT1DTPFDB_TOTAL_OID_LENGTH;
    *var_len = sizeof(long);    /* default to 'long' results */

    return MATCH_SUCCEEDED;
}

/*
 * var_dot1dTpFdbTable():
 *  This function contains Multi-Index tables
 *   The workings of this are basically the same as for var_dot1dTp above.
 */
unsigned char *
var_dot1dTpFdbTable(struct variable *vp,
            oid     *name,
            size_t  *length,
            int     exact,
            size_t  *var_len,
            WriteMethod **write_method)
{
    /* variables we may use later */
    static long long_ret;
    int count;
    static unsigned char string[SPRINT_MAX_LEN];

    /* Validate */
    if (header_dot1dTpFdbTable(vp,name,length,exact,var_len,write_method)
                                                == MATCH_FAILED ) {
        AGENT_EXIT;
        return NULL;
    }

    /* Start the MIB assignments */
    dot1dtp_parms->index = tp_entry_num;
    switch(vp->magic) {
        case DOT1DTPFDBADDRESS:
            *var_len = MAC_ADDRESS_LEN;
            for (count = 0; count < *var_len; count++) {
                string[count] = (unsigned char)name[DOT1DTPFDB_NAME_LENGTH+count];
            }
            return (u_char *) string;

        case DOT1DTPFDBPORT:
            dot1dtp_parms->group_obj = DOT1D_TPFDBPORT;
            break;

        case DOT1DTPFDBSTATUS:
            dot1dtp_parms->group_obj = DOT1D_TPFDBSTATUS;
            break;

        default:
            ERROR_MSG("Undefined MIB Reference in dot1dTp");
    }

    /* Callback */
    dot1dtp_parms->callback_rw = CALLBACK_ACT_READ;
    dot1dtp_parms->index = tp_entry_num;
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                        SNMPD_CALLBACK_BRIDGE,
                        dot1dtp_parms);
    
    long_ret = (long)dot1dtp_parms->value;
    return (unsigned char *)&long_ret;
}

static int
header_dot1dTpPortEntry(struct variable *vp,
           oid *name,
           size_t *length,
           int exact,
           size_t *var_len,
           WriteMethod **write_method)
{
#define DOT1DTPPORTENTRY_NAME_LENGTH    11
    oid newname[MAX_OID_LEN];
    int i, result;
    int found = FALSE;
    int port_number;

    memcpy( (char *)newname,(char *)vp->name, (int)vp->namelen * sizeof(oid));

    for(i=1; i <= interface_count; i++) {
        newname[DOT1DTPPORTENTRY_NAME_LENGTH] = (oid)i;
        result = snmp_oid_compare(
                name, *length, newname, (int)vp->namelen + 1);
        if ((exact && (result == 0)) || (!exact && (result < 0)))  {
            found = TRUE;
            port_number = i;
            break;
        }
    }

    /*TODO: Get AGG Port info? */
    if ( found == FALSE ) {
        DEBUGMSGTL(("dot1dTp", "... index out of range\n"));
        return MATCH_FAILED;
    }

    memcpy( (char *)name,(char *)newname, ((int)vp->namelen + 1) * sizeof(oid));
    *length = vp->namelen + 1;
    *write_method = 0;
    *var_len = sizeof(long);

    return port_number;
}

/*
 * var_dot1dTpPortTable():
 *  This function contains Multi-Index tables
 *   The workings of this are basically the same as for var_dot1dTp above.
 */
unsigned char *
var_dot1dTpPortTable(struct variable *vp,
            oid     *name,
            size_t  *length,
            int     exact,
            size_t  *var_len,
            WriteMethod **write_method)
{
    /* variables we may use later */
    static long long_ret;

    /* Simple table validation */
    snmp_mib_setInterfaceCount();
    if ((dot1dtp_parms->index = header_dot1dTpPortEntry(vp,name,length,exact,
                var_len,write_method)) == MATCH_FAILED ) {
        AGENT_EXIT;
        return NULL;
    }

    switch(vp->magic) {
        case DOT1DTPPORT:
            dot1dtp_parms->group_obj = DOT1D_TPPORT;
            break;
        case DOT1DTPPORTMAXINFO:
            dot1dtp_parms->group_obj = DOT1D_TPPORTMAXINFO;
            break;
        case DOT1DTPPORTINFRAMES:
            dot1dtp_parms->group_obj = DOT1D_TPPORTINFRAMES;
            break;
        case DOT1DTPPORTOUTFRAMES:
            dot1dtp_parms->group_obj = DOT1D_TPPORTOUTFRAMES;
            break;
        case DOT1DTPPORTINDISCARDS:
            dot1dtp_parms->group_obj = DOT1D_TPPORTINDISCARDS;
            break;
        default:
            AGENT_EXIT;
            return NULL;
    }

    /* Callback */
    dot1dtp_parms->callback_rw = CALLBACK_ACT_READ;
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                        SNMPD_CALLBACK_BRIDGE,
                        dot1dtp_parms);
    
    /* Give the value back */
    long_ret = (long)dot1dtp_parms->value;
    return (unsigned char *)&long_ret;
}

int
write_dot1dTpAgingTime(int      action,
            u_char   *var_val,
            u_char   var_val_type,
            size_t   var_val_len,
            u_char   *statP,
            oid      *name,
            size_t   name_len)
{
    long            value;
    value = *(long *) var_val;

    switch (action) {
    case RESERVE1:
        if (var_val_type != ASN_INTEGER) {
            fprintf(stderr, "write to dot1dTp not ASN_INTEGER\n");
            return SNMP_ERR_WRONGTYPE;
        }

        if (var_val_len > sizeof(long)) {
            fprintf(stderr, "write to dot1dTp: bad length\n");
            return SNMP_ERR_WRONGLENGTH;
        }

        if (value < C_L2_AGINGTIME_MIN || 
                value > C_L2_AGINGTIME_MAX) {
            return SNMP_ERR_WRONGVALUE;
        }
        break;

    case RESERVE2:
        break;

    case FREE:
        break;

    case ACTION:
        break;

    case UNDO:
        break;

    case COMMIT:
        /* We have enough info, so we can do a write now */
        dot1dtp_parms->value = value;
        AGENT_DBG(("Aging timer value: %d", dot1dtp_parms->value));
        dot1dtp_parms->group_obj = DOT1D_TPAGINGTIME;
        dot1dtp_parms->callback_rw = CALLBACK_ACT_WRITE;
        snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                            SNMPD_CALLBACK_BRIDGE,
                            dot1dtp_parms);
        if (dot1dtp_parms->rv == MIB_STATUS_ERROR) {
            AGENT_EXIT;
            return SNMP_ERR_COMMITFAILED;
        }
        break;
    }
    
    AGENT_EXIT;
    return SNMP_ERR_NOERROR;
}

