/*
 * $Id: dot1qVlan.c,v 1.1.2.8 2006/11/06 04:17:20 bobyu Exp $
 */

/* This should always be included first before anything else */
#include <config.h>

/* minimal include directives */
#include "mibincl.h"
#include "mibgroup/util_funcs.h"

#include "mibgroup/mibII/dot1qVlan.h"
#include <ucd-snmp/mibgroup/mibII/bridge_mib.h>

/* Argument to be passed with the callback function */
static bridge_parameters_t *dot1qvlan_parms = NULL;

/* Path to this MIB WRT the global MIB tree */
oid dot1qVlan_variables_oid[] = { 1,3,6,1,2,1,17,7,1,4 };

/* We can have only one VLAN under creation */
static qbridge_vlan_t *creation_row = NULL;

/* Status of the row under consideration */
static int curr_vid_index;

/* Variables that this MIB supports */
struct variable4 dot1qVlan_variables[] = {
/*  magic number        , variable type , ro/rw , callback fn  , L, oidsuffix */
#define DOT1QVLANNUMDELETES        1
    {DOT1QVLANNUMDELETES,  ASN_COUNTER,  RONLY ,  var_dot1qVlan, 1,  { 1 }},
#define DOT1QVLANTIMEMARK       2 
    {DOT1QVLANTIMEMARK,  ASN_TIMETICKS, RONLY, var_dot1qVlanCurrentTable, 3, {2, 1, 1}},
#define DOT1QVLANINDEX        3
    {DOT1QVLANINDEX,  ASN_UNSIGNED,  RONLY,   var_dot1qVlanCurrentTable, 3,  { 2, 1, 2 }},
#define DOT1QVLANFDBID        4
    {DOT1QVLANFDBID,  ASN_UNSIGNED,  RONLY,   var_dot1qVlanCurrentTable, 3,  { 2, 1, 3 }},
#define DOT1QVLANCURRENTEGRESSPORTS        5
    {DOT1QVLANCURRENTEGRESSPORTS,  ASN_OCTET_STR,  RONLY,   var_dot1qVlanCurrentTable, 3,  { 2, 1, 4 }},
#define DOT1QVLANCURRENTUNTAGGEDPORTS        6
    {DOT1QVLANCURRENTUNTAGGEDPORTS,  ASN_OCTET_STR,  RONLY,   var_dot1qVlanCurrentTable, 3,  { 2, 1, 5 }},
#define DOT1QVLANSTATUS        7
    {DOT1QVLANSTATUS,  ASN_INTEGER,  RONLY,   var_dot1qVlanCurrentTable, 3,  { 2, 1, 6 }},
#define DOT1QVLANCREATIONTIME        8
    {DOT1QVLANCREATIONTIME,  ASN_TIMETICKS,  RONLY,   var_dot1qVlanCurrentTable, 3,  { 2, 1, 7 }},
#define DOT1QVLANSTATICNAME        9
    {DOT1QVLANSTATICNAME,  ASN_OCTET_STR,  RWRITE,  var_dot1qVlanStaticTable, 3,  { 3, 1, 1 }},
#define DOT1QVLANSTATICEGRESSPORTS        10
    {DOT1QVLANSTATICEGRESSPORTS,  ASN_OCTET_STR,  RWRITE,  var_dot1qVlanStaticTable, 3,  { 3, 1, 2 }},
#define DOT1QVLANFORBIDDENEGRESSPORTS        11
    {DOT1QVLANFORBIDDENEGRESSPORTS,  ASN_OCTET_STR,  RWRITE,  var_dot1qVlanStaticTable, 3,  { 3, 1, 3 }},
#define DOT1QVLANSTATICUNTAGGEDPORTS        12
    {DOT1QVLANSTATICUNTAGGEDPORTS,  ASN_OCTET_STR,  RWRITE,  var_dot1qVlanStaticTable, 3,  { 3, 1, 4 }},
#define DOT1QVLANSTATICROWSTATUS        13
    {DOT1QVLANSTATICROWSTATUS,  ASN_INTEGER,  RWRITE,  var_dot1qVlanStaticTable, 3,  { 3, 1, 5 }},
#define DOT1QNEXTFREELOCALVLANINDEX        14
    {DOT1QNEXTFREELOCALVLANINDEX,  ASN_INTEGER,  RONLY ,  var_dot1qVlan, 1,  { 4 }},
#define DOT1QPVID        15
    {DOT1QPVID,  ASN_UNSIGNED,  RWRITE,  var_dot1qPortVlanTable, 3,  { 5, 1, 1 }},
#define DOT1QPORTACCEPTABLEFRAMETYPES        16
    {DOT1QPORTACCEPTABLEFRAMETYPES,  ASN_INTEGER,  RWRITE,  var_dot1qPortVlanTable, 3,  { 5, 1, 2 }},
#define DOT1QPORTINGRESSFILTERING        17
    {DOT1QPORTINGRESSFILTERING,  ASN_INTEGER,  RWRITE,  var_dot1qPortVlanTable, 3,  { 5, 1, 3 }},
#define DOT1QPORTGVRPSTATUS        18
    {DOT1QPORTGVRPSTATUS,  ASN_INTEGER,  RWRITE,  var_dot1qPortVlanTable, 3,  { 5, 1, 4 }},
#define DOT1QPORTGVRPFAILEDREGISTRATIONS        19
    {DOT1QPORTGVRPFAILEDREGISTRATIONS,  ASN_COUNTER,  RONLY,   var_dot1qPortVlanTable, 3,  { 5, 1, 5 }},
#define DOT1QPORTGVRPLASTPDUORIGIN        20
    {DOT1QPORTGVRPLASTPDUORIGIN,  ASN_OCTET_STR,  RONLY,   var_dot1qPortVlanTable, 3,  { 5, 1, 6 }},
#define DOT1QTPVLANPORTINFRAMES        21
    {DOT1QTPVLANPORTINFRAMES,  ASN_COUNTER,  RONLY,   var_dot1qPortVlanStatisticsTable, 3,  { 6, 1, 1 }},
#define DOT1QTPVLANPORTOUTFRAMES        22
    {DOT1QTPVLANPORTOUTFRAMES,  ASN_COUNTER,  RONLY,   var_dot1qPortVlanStatisticsTable, 3,  { 6, 1, 2 }},
#define DOT1QTPVLANPORTINDISCARDS        23
    {DOT1QTPVLANPORTINDISCARDS,  ASN_COUNTER,  RONLY,   var_dot1qPortVlanStatisticsTable, 3,  { 6, 1, 3 }},
#define DOT1QTPVLANPORTINOVERFLOWFRAMES        24
    {DOT1QTPVLANPORTINOVERFLOWFRAMES,  ASN_COUNTER,  RONLY,   var_dot1qPortVlanStatisticsTable, 3,  { 6, 1, 4 }},
#define DOT1QTPVLANPORTOUTOVERFLOWFRAMES        25
    {DOT1QTPVLANPORTOUTOVERFLOWFRAMES,  ASN_COUNTER,  RONLY,   var_dot1qPortVlanStatisticsTable, 3,  { 6, 1, 5 }},
#define DOT1QTPVLANPORTINOVERFLOWDISCARDS        26
    {DOT1QTPVLANPORTINOVERFLOWDISCARDS,  ASN_COUNTER,  RONLY,   var_dot1qPortVlanStatisticsTable, 3,  { 6, 1, 6 }},
#define DOT1QTPVLANPORTHCINFRAMES        27
    {DOT1QTPVLANPORTHCINFRAMES,  ASN_COUNTER64,  RONLY,   var_dot1qPortVlanHCStatisticsTable, 3,  { 7, 1, 1 }},
#define DOT1QTPVLANPORTHCOUTFRAMES        28
    {DOT1QTPVLANPORTHCOUTFRAMES,  ASN_COUNTER64,  RONLY,   var_dot1qPortVlanHCStatisticsTable, 3,  { 7, 1, 2 }},
#define DOT1QTPVLANPORTHCINDISCARDS        29
    {DOT1QTPVLANPORTHCINDISCARDS,  ASN_COUNTER64,  RONLY,   var_dot1qPortVlanHCStatisticsTable, 3,  { 7, 1, 3 }},
#define DOT1QCONSTRAINTVLAN        30
    {DOT1QCONSTRAINTVLAN,  ASN_UNSIGNED,  RONLY,   var_dot1qLearningConstraintsTable, 3,  { 8, 1, 1 }},
#define DOT1QCONSTRAINTSET        31
    {DOT1QCONSTRAINTSET,  ASN_INTEGER,  RONLY,   var_dot1qLearningConstraintsTable, 3,  { 8, 1, 2 }},
#define DOT1QCONSTRAINTTYPE        32
    {DOT1QCONSTRAINTTYPE,  ASN_INTEGER,  RWRITE,  var_dot1qLearningConstraintsTable, 3,  { 8, 1, 3 }},
#define DOT1QCONSTRAINTSTATUS        33
    {DOT1QCONSTRAINTSTATUS,  ASN_INTEGER,  RWRITE,  var_dot1qLearningConstraintsTable, 3,  { 8, 1, 4 }},
#define DOT1QCONSTRAINTSETDEFAULT        34
    {DOT1QCONSTRAINTSETDEFAULT,  ASN_INTEGER,  RWRITE,  var_dot1qVlan, 1,  { 9 }},
#define DOT1QCONSTRAINTTYPEDEFAULT        35
    {DOT1QCONSTRAINTTYPEDEFAULT,  ASN_INTEGER,  RWRITE,  var_dot1qVlan, 1,  { 10 }},
};

/** Initializes the dot1qVlan module */
void
init_dot1qVlan(void)
{
    AGENT_ENTER;

    /* BRCM SNMP init */
    dot1qvlan_parms = (bridge_parameters_t *) \
                sal_alloc(sizeof(bridge_parameters_t), MIBII_DOT1QVLAN_MIB);

    if (dot1qvlan_parms != NULL) {
        /* register ourselves with the agent to handle our mib tree */
        REGISTER_MIB(MIBII_DOT1QVLAN_MIB, dot1qVlan_variables, variable4,
                   dot1qVlan_variables_oid);

        dot1qvlan_parms->group = BRIDGE_DOT1QVLAN; 
    }

    AGENT_EXIT;
}

/*
 * 802.1Q VLAN Global Variables
 */
unsigned char *
var_dot1qVlan(struct variable *vp, 
                oid     *name, 
                size_t  *length, 
                int     exact, 
                size_t  *var_len, 
                WriteMethod **write_method)
{
    AGENT_ENTER;
    static u_long ulong_ret;

    if (header_generic(vp,name,length,exact,var_len,write_method)
                                  == MATCH_FAILED ) {
        AGENT_EXIT;
        return NULL;
    }

    switch(vp->magic) {
        case DOT1QVLANNUMDELETES:
            dot1qvlan_parms->group_obj = DOT1Q_VLANNUMDELETES;
            break;
        case DOT1QNEXTFREELOCALVLANINDEX:
            dot1qvlan_parms->group_obj = DOT1Q_NEXTFREELOCALVLANINDEX;
            break;
        case DOT1QCONSTRAINTSETDEFAULT:
            *write_method = write_dot1qConstraintSetDefault;
            dot1qvlan_parms->group_obj = DOT1Q_CONSTRAINTSETDEFAULT;
            break;
        case DOT1QCONSTRAINTTYPEDEFAULT:
            *write_method = write_dot1qConstraintTypeDefault;
            dot1qvlan_parms->group_obj = DOT1Q_CONSTRAINTTYPEDEFAULT;
            break;
        default:
            AGENT_EXIT;
            return NULL;
    }

    /* Callback */
    dot1qvlan_parms->callback_rw = CALLBACK_ACT_READ;
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                        SNMPD_CALLBACK_BRIDGE,
                        dot1qvlan_parms);
    
    /* Give the value back */
    ulong_ret = (u_long)dot1qvlan_parms->value;
    return (unsigned char *)&ulong_ret;
}


/*
 *  This table is not supported since it uses the timefiler
 *  concept for getting the VLAN status.
 */
unsigned char *
var_dot1qVlanCurrentTable(struct variable *vp,
            oid     *name,
            size_t  *length,
            int     exact,
            size_t  *var_len,
            WriteMethod **write_method)
{
    AGENT_ENTER;
    static long long_ret;
    static unsigned char string[SPRINT_MAX_LEN];

    /* Proceed only if VLAN mode is 802.1Q */
    if (snmp_mib_getVlanMode() != MIB_VLAN_MODE_TAGGED) {
        AGENT_EXIT;
        return NULL;
    }

    /* For now, we just return. */
    AGENT_EXIT;
    return NULL;

    if (header_simple_table(vp,name,length,exact,var_len,write_method, 10)
                                               == MATCH_FAILED ) {
        AGENT_EXIT;
        return NULL;
    }

  /* 
   * Start the MIB assignments
   */
    switch(vp->magic) {
    case DOT1QVLANTIMEMARK:
        dot1qvlan_parms->group_obj = DOT1Q_VLANTIMEMARK;
        break;
    case DOT1QVLANINDEX:
        dot1qvlan_parms->group_obj = DOT1Q_VLANINDEX;
        break;
    case DOT1QVLANFDBID:
        dot1qvlan_parms->group_obj = DOT1Q_VLANFDBID;
        break;
    case DOT1QVLANCURRENTEGRESSPORTS:
        dot1qvlan_parms->group_obj = DOT1Q_VLANCURRENTEGRESSPORTS;
        break;
    case DOT1QVLANCURRENTUNTAGGEDPORTS:
        dot1qvlan_parms->group_obj = DOT1Q_VLANCURRENTUNTAGGEDPORTS;
        break;
    case DOT1QVLANSTATUS:
        dot1qvlan_parms->group_obj = DOT1Q_VLANSTATUS;
        break;
    case DOT1QVLANCREATIONTIME:
        dot1qvlan_parms->group_obj = DOT1Q_VLANCREATIONTIME;
        break;
    default:
       AGENT_EXIT;
       return NULL;
    }

    /* Callback */
    dot1qvlan_parms->callback_rw = CALLBACK_ACT_READ;
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                        SNMPD_CALLBACK_BRIDGE,
                        dot1qvlan_parms);
    
    /* Give the value back */
    long_ret = (long)dot1qvlan_parms->value;
    return (unsigned char *)&long_ret;
}

/*
 * Validation function for 802.1q VLAN Static table.
 * This is basically a list of VLANs that are 
 * created and saved in persistence
 */
static int
header_dot1qVlanStaticTable(struct variable *vp,
           oid *name,
           size_t *length,
           int exact,
           size_t *var_len,
           WriteMethod **write_method)
{
#define DOT1Q_VLAN_STATIC_NAME_LEN  13
#define DOT1Q_VLAN_STATIC_TOTL_LEN  14

    AGENT_ENTER;
    oid newname[MAX_OID_LEN];
    oid lowest[MAX_OID_LEN];
    int i=0, result, found = FALSE;
    int vlan_id, vlan_count; 

    /* In case we need to create new row... */
    *write_method = write_dot1qVlanStaticEntry;
    
    /* Get count of number of VLANs on the switch */
    dot1qvlan_parms->callback_rw = CALLBACK_ACT_READ;
    dot1qvlan_parms->group_obj = DOT1Q_VLANCOUNT;
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION,
                        SNMPD_CALLBACK_BRIDGE,
                        dot1qvlan_parms);
    vlan_count = dot1qvlan_parms->value;

    memcpy( (char *)newname,(char *)vp->name, (int)vp->namelen * sizeof(oid));
    /* VLAN under creation because of createAndWait */
    if (creation_row) {
        if (creation_row->rowStatus != MIB_ROWSTATUS_ACTIVE) {
            /* New row */
            i = -1;
        }
    }

    for (; i < vlan_count ; i++) {
        if (i == -1) { /* First, check VLAN under creation*/
            vlan_id = creation_row->vid;
        } else {
            dot1qvlan_parms->index = i;
            dot1qvlan_parms->group_obj = DOT1Q_VLANID;
            snmp_call_callbacks(SNMP_CALLBACK_APPLICATION,
                                SNMPD_CALLBACK_BRIDGE,
                                dot1qvlan_parms);
            vlan_id = dot1qvlan_parms->value;
            AGENT_DBG(("VLAN ID = %d", vlan_id));
        }

        if (vlan_id < 1 || vlan_id > 4094) {
            continue;
        }

        newname[DOT1Q_VLAN_STATIC_NAME_LEN] = (oid) vlan_id ;
        if (exact) {
            if (snmp_oid_compare(name, *length, newname, 
                        DOT1Q_VLAN_STATIC_TOTL_LEN) == 0) {
                memcpy(lowest, newname, 
                                DOT1Q_VLAN_STATIC_TOTL_LEN * sizeof(oid));
                found = TRUE;
                curr_vid_index = i;
                AGENT_DBG(("Index exact %d", i));
                break;
            }
        } else {
            if ((snmp_oid_compare(newname, DOT1Q_VLAN_STATIC_TOTL_LEN, 
                   name, *length) > 0) && (found == FALSE || 
                       snmp_oid_compare(newname, 
                       DOT1Q_VLAN_STATIC_TOTL_LEN, lowest, 
                           DOT1Q_VLAN_STATIC_TOTL_LEN) < 0))  {
                memcpy((char *) lowest, (char *) newname,
                       DOT1Q_VLAN_STATIC_TOTL_LEN * sizeof(oid));
                AGENT_DBG(("Index %d", i));
                curr_vid_index = i;
                found = TRUE;
            }
        }
    }

    if (!found) {
        AGENT_EXIT;
        return MATCH_FAILED;
    }

    memcpy( (char *)name,(char *)lowest, ((int) vp->namelen +1) * sizeof(oid));
    *length = vp->namelen + 1;
    *var_len = sizeof(long);    /* default to 'long' results */

    return MATCH_SUCCEEDED;
}

/*
 * var_dot1qVlanStaticTable():
 *  This function contains Multi-Index tables
 *   The workings of this are basically the same as for var_dot1qVlan above.
 */
unsigned char *
var_dot1qVlanStaticTable(struct variable *vp,
            oid     *name,
            size_t  *length,
            int     exact,
            size_t  *var_len,
            WriteMethod **write_method)
{
    AGENT_ENTER;
    static long long_ret;
    static unsigned char string[SPRINT_MAX_LEN];

    /* Proceed only if VLAN mode is 802.1Q */
    if (snmp_mib_getVlanMode() != MIB_VLAN_MODE_TAGGED) {
        AGENT_EXIT;
        return NULL;
    }

    snmp_mib_setInterfaceCount();
    if (header_dot1qVlanStaticTable(vp,name,length,exact,var_len,write_method)
                                                == MATCH_FAILED ) {
        AGENT_EXIT;
        return NULL;    
    }

    dot1qvlan_parms->index = curr_vid_index;
    switch(vp->magic) {
        case DOT1QVLANSTATICNAME:
            if (curr_vid_index == -1) {
                strcpy (dot1qvlan_parms->buf, creation_row->vlanName);
                dot1qvlan_parms->value = strlen(creation_row->vlanName);
            } else {
                dot1qvlan_parms->group_obj = DOT1Q_VLANSTATICNAME;
            }
            break;
        
        case DOT1QVLANSTATICEGRESSPORTS:
            if (curr_vid_index == -1) {
                dot1qvlan_parms->value = interface_count/INTERFACE_BYTE_COUNT;             
                memcpy (dot1qvlan_parms->buf, creation_row->tports, 
                                               dot1qvlan_parms->value);
            } else {
                dot1qvlan_parms->group_obj = DOT1Q_VLANSTATICEGRESSPORTS;
            }
            break;
        
        case DOT1QVLANFORBIDDENEGRESSPORTS:
            if (curr_vid_index == -1) {
                dot1qvlan_parms->value = interface_count/INTERFACE_BYTE_COUNT;             
                memset (dot1qvlan_parms->buf, 0, sizeof(dot1qvlan_parms->buf));
            } else {
                dot1qvlan_parms->group_obj = DOT1Q_VLANFORBIDDENEGRESSPORTS;
            }
            break;
        
        case DOT1QVLANSTATICUNTAGGEDPORTS:
            if (curr_vid_index == -1) {
                dot1qvlan_parms->value = interface_count/INTERFACE_BYTE_COUNT;             
                memcpy (dot1qvlan_parms->buf, creation_row->utports, 
                                                dot1qvlan_parms->value);
            } else {
                dot1qvlan_parms->group_obj = DOT1Q_VLANSTATICUNTAGGEDPORTS;
            }
            break;
        
        case DOT1QVLANSTATICROWSTATUS:
            if (curr_vid_index == -1) {
                dot1qvlan_parms->value = creation_row->rowStatus;
            } else {
                dot1qvlan_parms->group_obj = DOT1Q_VLANSTATICROWSTATUS;
            }
            break;

        default:
            AGENT_EXIT;
            return NULL;
    }

    if (curr_vid_index != -1) {
        dot1qvlan_parms->callback_rw = CALLBACK_ACT_READ;
        snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                            SNMPD_CALLBACK_BRIDGE,
                            dot1qvlan_parms);
    }

    switch(vp->magic) {
        case DOT1QVLANSTATICNAME:
        case DOT1QVLANSTATICEGRESSPORTS:
        case DOT1QVLANFORBIDDENEGRESSPORTS:
        case DOT1QVLANSTATICUNTAGGEDPORTS:
            *var_len = dot1qvlan_parms->value;
            memcpy (string, dot1qvlan_parms->buf, dot1qvlan_parms->value);
            return (unsigned char *) string;
    }
    
    /* Give the value back */
    long_ret = (long)dot1qvlan_parms->value;
    return (unsigned char *)&long_ret;
}

static int
header_dot1qPortVlanEntry(struct variable *vp,
           oid *name,
           size_t *length,
           int exact,
           size_t *var_len,
           WriteMethod **write_method)
{
#define DOT1Q_PORT_VLAN_ENTRY 13
    oid newname[MAX_OID_LEN];
    int i, result;
    int port_number;
    int found = FALSE;
      
    memcpy( (char *)newname,(char *)vp->name, (int)vp->namelen * sizeof(oid));
    
    for(i = 1; i <= interface_count; i++) {
        newname[DOT1Q_PORT_VLAN_ENTRY] = (oid)i;
        result = snmp_oid_compare(
                    name, *length, newname, (int)vp->namelen + 1);
        if ((exact && (result == 0)) || (!exact && (result < 0))) {
            found = TRUE;
            port_number = i;
            break;
        }
    }

    if ( found == FALSE ) {
        AGENT_EXIT;
        return MATCH_FAILED;
    }

    memcpy( (char *)name,(char *)newname, ((int)vp->namelen + 1) * sizeof(oid));

    *length = vp->namelen + 1;
    *write_method = 0;
    *var_len = sizeof(long);    /* default to 'long' results */

    return port_number;
}

/*
 * var_dot1qPortVlanTable():
 *  This function contains Multi-Index tables
 *   The workings of this are basically the same as for var_dot1qVlan above.
 */
unsigned char *
var_dot1qPortVlanTable(struct variable *vp,
            oid     *name,
            size_t  *length,
            int     exact,
            size_t  *var_len,
            WriteMethod **write_method)
{
    AGENT_ENTER;
    static long long_ret;
    static unsigned char string[SPRINT_MAX_LEN];

    /* Proceed only if VLAN mode is 802.1Q */
    if (snmp_mib_getVlanMode() != MIB_VLAN_MODE_TAGGED) {
        AGENT_EXIT;
        return NULL;
    }

    snmp_mib_setInterfaceCount();
    if ((dot1qvlan_parms->index = header_dot1qPortVlanEntry(vp,name,length,
                               exact,var_len,write_method)) == MATCH_FAILED ) {
        AGENT_EXIT;
        return NULL;
    }

    switch(vp->magic) {
        case DOT1QPVID:
            *write_method = write_dot1qPvid;
            dot1qvlan_parms->group_obj = DOT1Q_PVID;
            break;
        case DOT1QPORTACCEPTABLEFRAMETYPES:
            *write_method = write_dot1qPortAcceptableFrameTypes;
            dot1qvlan_parms->group_obj = DOT1Q_PORTACCEPTABLEFRAMETYPES;
            break;
        case DOT1QPORTINGRESSFILTERING:
            *write_method = write_dot1qPortIngressFiltering;
            dot1qvlan_parms->group_obj = DOT1Q_PORTINGRESSFILTERING;
            break;
        case DOT1QPORTGVRPSTATUS:
            *write_method = write_dot1qPortGvrpStatus;
            dot1qvlan_parms->group_obj = DOT1Q_PORTGVRPSTATUS;
            break;
        case DOT1QPORTGVRPFAILEDREGISTRATIONS:
            dot1qvlan_parms->group_obj = DOT1Q_PORTGVRPFAILEDREGISTRATIONS;
            break;
        case DOT1QPORTGVRPLASTPDUORIGIN:
            dot1qvlan_parms->group_obj = DOT1Q_PORTGVRPLASTPDUORIGIN;
            break;
        default:
            AGENT_EXIT;
            return NULL;
    }

    dot1qvlan_parms->callback_rw = CALLBACK_ACT_READ;
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                        SNMPD_CALLBACK_BRIDGE,
                        dot1qvlan_parms);
    
    switch(vp->magic) {
        case DOT1QPORTGVRPLASTPDUORIGIN:
            *var_len = dot1qvlan_parms->value;
            memcpy (string, dot1qvlan_parms->buf, *var_len);
            return (unsigned char *) string;
    }

    long_ret = (long)dot1qvlan_parms->value;
    return (unsigned char *)&long_ret;
}

static int
header_dot1qPortVlanStatisticsTable(struct variable *vp,
           oid *name,
           size_t *length,
           int exact,
           size_t *var_len,
           WriteMethod **write_method)
{
#define DOT1Q_PORT_VLANSTAT_ENTRY 13
#define DOT1Q_PORT_VLANSTAT_LEN 15
    AGENT_ENTER;

    oid newname[MAX_OID_LEN];
    int i, j, result, vlan_count = 0;
    oid lowest[MAX_OID_LEN];
    int found = FALSE;

    memcpy( (char *)newname,(char *)vp->name, (int)vp->namelen * sizeof(oid));
    lowest[0] = 99999;

    /* Get VLAN Count */
    dot1qvlan_parms->callback_rw = CALLBACK_ACT_READ;
    dot1qvlan_parms->group_obj = DOT1Q_VLANCOUNT;
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION,
                        SNMPD_CALLBACK_BRIDGE,
                        dot1qvlan_parms);
    vlan_count = dot1qvlan_parms->value;
    AGENT_DBG(("Vlan count %d", vlan_count));

    for(i = 1; i <= interface_count; i++) {
        newname[DOT1Q_PORT_VLANSTAT_ENTRY] = (oid)i;
        AGENT_DBG(("For port %d", i));
        for (j = 0; j < vlan_count; j++) {
            dot1qvlan_parms->index = j;
            dot1qvlan_parms->group_obj = DOT1Q_VLANID;
            snmp_call_callbacks(SNMP_CALLBACK_APPLICATION,
                                SNMPD_CALLBACK_BRIDGE,
                                dot1qvlan_parms);
            newname[DOT1Q_PORT_VLANSTAT_ENTRY+1] = (oid)dot1qvlan_parms->value;
            AGENT_DBG(("VLAN %d", newname[DOT1Q_PORT_VLANSTAT_ENTRY+1]));

            if (exact) {
                if (snmp_oid_compare(name, *length, newname, 
                            DOT1Q_PORT_VLANSTAT_LEN) == 0) {
                    memcpy((char *) lowest, (char *)newname, 
                                    DOT1Q_PORT_VLANSTAT_LEN * sizeof(oid));
                    found = TRUE;
                    break;
                }
            } else {
                if ((snmp_oid_compare(newname, DOT1Q_PORT_VLANSTAT_LEN, 
                       name, *length) > 0) && (found == FALSE || 
                           snmp_oid_compare(newname, 
                           DOT1Q_PORT_VLANSTAT_LEN, lowest, 
                               DOT1Q_PORT_VLANSTAT_LEN) < 0))  {
                    memcpy((char *) lowest, (char *) newname,
                           DOT1Q_PORT_VLANSTAT_LEN * sizeof(oid));
                    found = TRUE;
                }
            }
        }
    }

    if ( found == FALSE ) {
        AGENT_EXIT;
        return MATCH_FAILED;
    }

    memcpy( (char *)name,(char *)lowest, DOT1Q_PORT_VLANSTAT_LEN * sizeof(oid));
    AGENT_OID_PRINT(DOT1Q_PORT_VLANSTAT_LEN, name);

    *length = DOT1Q_PORT_VLANSTAT_LEN;
    *write_method = 0;
    *var_len = sizeof(long);    /* default to 'long' results */
    
    AGENT_EXIT;
    return MATCH_SUCCEEDED;
}

/*
 * This keeps a count 0f number of stats for a port on
 * a particular VLAN. Per VLAN port stats are not currently suppported.
 */
unsigned char *
var_dot1qPortVlanStatisticsTable(struct variable *vp,
            oid     *name,
            size_t  *length,
            int     exact,
            size_t  *var_len,
            WriteMethod **write_method)
{
    AGENT_ENTER;
    static u_long ulong_ret;

    /* Proceed only if VLAN mode is 802.1Q */
    if (snmp_mib_getVlanMode() != MIB_VLAN_MODE_TAGGED) {
        AGENT_EXIT;
        return NULL;
    }

    /* Get the ports, and for each port get VLANs it is part of */
    snmp_mib_setInterfaceCount();
    if (header_dot1qPortVlanStatisticsTable(vp,name,length,exact,var_len,
                                           write_method) == MATCH_FAILED ) {
        AGENT_EXIT;
        return NULL;
    }

    dot1qvlan_parms->index = (int) name[13]; /* Port ID */
    dot1qvlan_parms->value = (int) name[14]; /* VLAN ID */
    switch(vp->magic) {
        case DOT1QTPVLANPORTINFRAMES:
            dot1qvlan_parms->group_obj = DOT1Q_TPVLANPORTINFRAMES;
            break;
        case DOT1QTPVLANPORTOUTFRAMES:
            dot1qvlan_parms->group_obj = DOT1Q_TPVLANPORTOUTFRAMES;
            break;
        case DOT1QTPVLANPORTINDISCARDS:
            dot1qvlan_parms->group_obj = DOT1Q_TPVLANPORTINDISCARDS;
            break;
        case DOT1QTPVLANPORTINOVERFLOWFRAMES:
            dot1qvlan_parms->group_obj = DOT1Q_TPVLANPORTINOVERFLOWFRAMES;
            break;
        case DOT1QTPVLANPORTOUTOVERFLOWFRAMES:
            dot1qvlan_parms->group_obj = DOT1Q_TPVLANPORTOUTOVERFLOWFRAMES;
            break;
        case DOT1QTPVLANPORTINOVERFLOWDISCARDS:
            dot1qvlan_parms->group_obj = DOT1Q_TPVLANPORTINOVERFLOWDISCARDS;
            break;
        default:
            AGENT_EXIT;
            return NULL;
    }
    
    dot1qvlan_parms->callback_rw = CALLBACK_ACT_READ;
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                        SNMPD_CALLBACK_BRIDGE,
                        dot1qvlan_parms);
    
    /* Give the value back */
    ulong_ret = (u_long)dot1qvlan_parms->value;
    return (unsigned char *)&ulong_ret;
}

/*
 * Since no high capacity interfaces are present, we do not
 * support this for now.
 */
unsigned char *
var_dot1qPortVlanHCStatisticsTable(struct variable *vp,
            oid     *name,
            size_t  *length,
            int     exact,
            size_t  *var_len,
            WriteMethod **write_method)
{
    /* variables we may use later */
    static u_long ulong_ret;

    /* Proceed only if VLAN mode is 802.1Q */
    if (snmp_mib_getVlanMode() != MIB_VLAN_MODE_TAGGED) {
        AGENT_EXIT;
        return NULL;
    }

    /* For now, not supported */
    return NULL;

    snmp_mib_setInterfaceCount();
    if (header_dot1qPortVlanStatisticsTable(vp,name,length,exact,
                                    var_len,write_method) == MATCH_FAILED ) {
        AGENT_EXIT;
        return NULL;
    }

    AGENT_OID_PRINT(*length, name);
    
    dot1qvlan_parms->index = (int) name[13];
    dot1qvlan_parms->value = (int) name[14];
    AGENT_DBG (("header and value %d, %d",
                  dot1qvlan_parms->index, dot1qvlan_parms->value));
    switch(vp->magic) {
        case DOT1QTPVLANPORTHCINFRAMES:
            dot1qvlan_parms->group_obj = DOT1Q_TPVLANPORTHCINFRAMES;
            break;
        case DOT1QTPVLANPORTHCOUTFRAMES:
            dot1qvlan_parms->group_obj = DOT1Q_TPVLANPORTHCOUTFRAMES;
            break;
        case DOT1QTPVLANPORTHCINDISCARDS:
            dot1qvlan_parms->group_obj = DOT1Q_TPVLANPORTHCINDISCARDS;
            break;
        default:
            AGENT_EXIT;
            return NULL;
    }

    dot1qvlan_parms->callback_rw = CALLBACK_ACT_READ;
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                        SNMPD_CALLBACK_BRIDGE,
                        dot1qvlan_parms);
    
    /* Give the value back */
    ulong_ret = (u_long)dot1qvlan_parms->value;
    AGENT_EXIT;
    return (unsigned char *)&ulong_ret;
}

/*
 * Again, this is not supported
 */
unsigned char *
var_dot1qLearningConstraintsTable(struct variable *vp,
            oid     *name,
            size_t  *length,
            int     exact,
            size_t  *var_len,
            WriteMethod **write_method)
{
    /* variables we may use later */
    static long long_ret;

    /* Proceed only if VLAN mode is 802.1Q */
    if (snmp_mib_getVlanMode() != MIB_VLAN_MODE_TAGGED) {
        AGENT_EXIT;
        return NULL;
    }

    /* for now, no support */
    return NULL;

    if (header_simple_table(vp,name,length,exact,var_len,write_method, 10)
                                                == MATCH_FAILED ) {
        AGENT_EXIT;
        return NULL;
    }

  /* 
   * Start the MIB assignments
   */
    switch(vp->magic) {
        case DOT1QCONSTRAINTVLAN:
            dot1qvlan_parms->group_obj = DOT1Q_CONSTRAINTVLAN;
            break;
        case DOT1QCONSTRAINTSET:
            dot1qvlan_parms->group_obj = DOT1Q_CONSTRAINTSET;
            break;
        case DOT1QCONSTRAINTTYPE:
            *write_method = write_dot1qConstraintType;
            dot1qvlan_parms->group_obj = DOT1Q_CONSTRAINTTYPE;
            break;
        case DOT1QCONSTRAINTSTATUS:
            *write_method = write_dot1qConstraintStatus;
            dot1qvlan_parms->group_obj = DOT1Q_CONSTRAINTSTATUS;
            break;
        default:
            AGENT_EXIT;
            return NULL;
    }

    dot1qvlan_parms->callback_rw = CALLBACK_ACT_READ;
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                        SNMPD_CALLBACK_BRIDGE,
                        dot1qvlan_parms);
    
    /* Give the value back */
    long_ret = (long)dot1qvlan_parms->value;
    return (unsigned char *)&long_ret;
}

/* Common read-create function for Static Vlan */
int
write_dot1qVlanStaticEntry(int action,
                          u_char * var_val,
                          u_char var_val_type,
                          size_t var_val_len,
                          u_char * statP, oid * name, size_t name_len)
{
    char            value;
    int             var, curr_status;
 
    /* variable being set */
    var = name[DOT1Q_VLAN_STATIC_NAME_LEN-1]+8;

    /* Get the status of the VLAN being set */
    dot1qvlan_parms->value = name[DOT1Q_VLAN_STATIC_NAME_LEN];
    dot1qvlan_parms->callback_rw = CALLBACK_ACT_READ;
    dot1qvlan_parms->group_obj = DOT1Q_VIDQUERY;
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION,
                        SNMPD_CALLBACK_BRIDGE,
                        dot1qvlan_parms);
    curr_status = dot1qvlan_parms->value;
   
    /* Since all variables are read create, we create row by default */
    if (creation_row == NULL) {
        creation_row = (qbridge_vlan_t *) \
                        sal_alloc(sizeof(qbridge_vlan_t), "1qVlanCreate");

        if (creation_row == NULL) {
            AGENT_EXIT;
            return SNMP_ERR_RESOURCEUNAVAILABLE;
        }

        /* cannot set active vlan status to anything other than 6 */
        if (curr_status == 1 && 
                    (var == DOT1QVLANSTATICROWSTATUS && value != 6)) {
            AGENT_EXIT;
            return SNMP_ERR_WRONGVALUE;
        }

        /* zero out all fields - first time only */
        memset(creation_row, 0, sizeof(qbridge_vlan_t));

        /* Get the vlan id from oid */
        creation_row->vid = 
                    (int)name[DOT1Q_VLAN_STATIC_NAME_LEN] ;

        /* Set the row status to notReady */
        creation_row->rowStatus = curr_status;
    } else if (creation_row != NULL) {
        if (name[DOT1Q_VLAN_STATIC_NAME_LEN] != creation_row->vid) {
            AGENT_DBG(("Create row VID = %d and curr vid = %d",
                        creation_row->vid, name[DOT1Q_VLAN_STATIC_NAME_LEN]));
            AGENT_EXIT;
            return SNMP_ERR_RESOURCEUNAVAILABLE;
        }
    }

    /* If already existing entry, get current values */
    if (curr_status == 1) {
        dot1qvlan_parms->value = name[DOT1Q_VLAN_STATIC_NAME_LEN];
        dot1qvlan_parms->group_obj = DOT1Q_VLANBULKGET;
        snmp_call_callbacks(SNMP_CALLBACK_APPLICATION,
                            SNMPD_CALLBACK_BRIDGE,
                            dot1qvlan_parms);
        memcpy (creation_row, 
                  &dot1qvlan_parms->group_param.dot1qvlan_param, 
                     sizeof(qbridge_vlan_t));
    }

    switch (action) {
    case RESERVE1:
        switch(var) {
            case DOT1QVLANSTATICNAME:
                    if (var_val_type != ASN_OCTET_STR) {
                        fprintf(stderr, 
                                "write to dot1qVlan not ASN_OCTET_STR\n");
                        return SNMP_ERR_WRONGTYPE;
                    }
                    if (var_val_len > sizeof(char) * C_VLAN_NAME_SIZE) {
                        fprintf(stderr, "write to dot1qVlan: bad length\n");
                        return SNMP_ERR_WRONGLENGTH;
                    }
                    break;

            case DOT1QVLANSTATICEGRESSPORTS:
                if (var_val_type != ASN_OCTET_STR) {
                    fprintf(stderr, "write to dot1qVlan not ASN_OCTET_STR\n");
                    return SNMP_ERR_WRONGTYPE;
                }
                
                if (var_val_len > (sizeof(char) * interface_count/INTERFACE_BYTE_COUNT)) {
                    fprintf(stderr, "write to dot1qVlan: bad length\n");
                    return SNMP_ERR_WRONGLENGTH;
                }
                break;

            case DOT1QVLANSTATICUNTAGGEDPORTS:
                if (var_val_type != ASN_OCTET_STR) {
                    fprintf(stderr, "write to dot1qVlan not ASN_OCTET_STR\n");
                    return SNMP_ERR_WRONGTYPE;
                }
                if (var_val_len > (sizeof(char) * interface_count/INTERFACE_BYTE_COUNT)) {
                    fprintf(stderr, "write to dot1qVlan: bad length\n");
                    return SNMP_ERR_WRONGLENGTH;
                }
                break;

            case DOT1QVLANSTATICROWSTATUS:
                if (var_val_type != ASN_INTEGER) {
                    fprintf(stderr, "write to dot1qVlan not ASN_INTEGER\n");
                    return SNMP_ERR_WRONGTYPE;
                }
                if (var_val_len > sizeof(long)) {
                    fprintf(stderr, "write to dot1qVlan: bad length\n");
                    return SNMP_ERR_WRONGLENGTH;
                }

                value = *(long *) var_val;
                if (value < 1 || value > 6) {
                    fprintf(stderr, "write to dot1qVlan: bad value\n");
                    return SNMP_ERR_WRONGVALUE;
                }
                break;
        }
        break;


    case RESERVE2:
        break;

    case FREE:
        if (creation_row)
            sal_free(creation_row);
        creation_row = NULL;
        break;

    case ACTION:
        break;

    case UNDO:
        /*
         * Back out any changes made in the ACTION case 
         */
        break;

    case COMMIT:
         switch(var) {
            case DOT1QVLANSTATICNAME:
                memcpy(creation_row->vlanName, var_val, var_val_len);
                creation_row->vlanName[var_val_len] = 0;
                break;

            case DOT1QVLANSTATICEGRESSPORTS:
                memcpy(creation_row->tports, var_val, interface_count/INTERFACE_BYTE_COUNT);
                break;

            case DOT1QVLANSTATICUNTAGGEDPORTS:
                memcpy(creation_row->utports, var_val, interface_count/INTERFACE_BYTE_COUNT);
                break;

            case DOT1QVLANSTATICROWSTATUS:
                creation_row->rowStatus = *(long *)var_val;
                break;
        }
        if (creation_row->rowStatus != MIB_ROWSTATUS_DESTROY&& 
                creation_row->rowStatus != MIB_ROWSTATUS_ACTIVE&
                    creation_row->rowStatus != MIB_ROWSTATUS_CREATEANDGO) {
            /* All these status do not go to backend */
            return SNMP_ERR_NOERROR;
        }
        memcpy (&dot1qvlan_parms->group_param.dot1qvlan_param,
                creation_row, sizeof(qbridge_vlan_t));
        dot1qvlan_parms->callback_rw = CALLBACK_ACT_WRITE;
        dot1qvlan_parms->group_obj = DOT1Q_VLANSTATICROWSTATUS;
        snmp_call_callbacks(SNMP_CALLBACK_APPLICATION,
                            SNMPD_CALLBACK_BRIDGE,
                            dot1qvlan_parms);
        if (dot1qvlan_parms->rv == MIB_STATUS_ERROR) {
            /* We do not free memory here. */
            AGENT_EXIT;
            return SNMP_ERR_COMMITFAILED;
        }

        if (creation_row) {
            sal_free(creation_row);
            creation_row = NULL;
        }
        break;
    }

    AGENT_EXIT;
    return SNMP_ERR_NOERROR;
}

int
write_dot1qPvid(int action,
                u_char * var_val,
                u_char var_val_type,
                size_t var_val_len,
                u_char * statP, oid * name, size_t name_len)
{
    u_long          value;
    
    dot1qvlan_parms->index = name[DOT1Q_PORT_VLAN_ENTRY];
    switch (action) {
    case RESERVE1:
        if (var_val_type != ASN_UNSIGNED) {
            fprintf(stderr, "write to dot1qVlan not ASN_UNSIGNED\n");
            return SNMP_ERR_WRONGTYPE;
        }
        if (var_val_len > sizeof(u_long)) {
            fprintf(stderr, "write to dot1qVlan: bad length\n");
            return SNMP_ERR_WRONGLENGTH;
        }
        value = *(u_long *) var_val;
        if (value < 1 || value > 4094) {
            fprintf(stderr, "write to dot1qVlan: bad value\n");
            return SNMP_ERR_WRONGVALUE;
        }
        break;

    case RESERVE2:
        break;

    case FREE:
        break;

    case ACTION:
        break;

    case UNDO:
        break;

    case COMMIT:
        dot1qvlan_parms->value = * (long *) var_val;
        dot1qvlan_parms->callback_rw = CALLBACK_ACT_WRITE;
        dot1qvlan_parms->group_obj = DOT1Q_PVID;
        snmp_call_callbacks(SNMP_CALLBACK_APPLICATION,
                                SNMPD_CALLBACK_BRIDGE,
                                dot1qvlan_parms);
        if (dot1qvlan_parms->rv == MIB_STATUS_ERROR) {
            AGENT_EXIT;
            return SNMP_ERR_COMMITFAILED;
        }
        break;
    }

    AGENT_EXIT;
    return SNMP_ERR_NOERROR;
}

int
write_dot1qPortAcceptableFrameTypes(int action,
                                    u_char * var_val,
                                    u_char var_val_type,
                                    size_t var_val_len,
                                    u_char * statP,
                                    oid * name, size_t name_len)
{
    long            value;
    dot1qvlan_parms->index = name[DOT1Q_PORT_VLAN_ENTRY];

    switch (action) {
    case RESERVE1:
        if (var_val_type != ASN_INTEGER) {
            fprintf(stderr, "write to dot1qVlan not ASN_INTEGER\n");
            return SNMP_ERR_WRONGTYPE;
        }
        if (var_val_len > sizeof(long)) {
            fprintf(stderr, "write to dot1qVlan: bad length\n");
            return SNMP_ERR_WRONGLENGTH;
        }
        value = *(u_long *) var_val;
        if (value != C_PORT_FRAMETYPE_ALL && 
                        value != C_PORT_FRAMETYPE_TAGGED) {
            fprintf(stderr, "write to dot1qVlan: bad value\n");
            return SNMP_ERR_WRONGVALUE;
        }
        break;

    case RESERVE2:
        break;

    case FREE:
        break;

    case ACTION:
        break;

    case UNDO:
        break;

    case COMMIT:
        dot1qvlan_parms->value = * (long *) var_val;
        dot1qvlan_parms->group_obj = DOT1Q_PORTACCEPTABLEFRAMETYPES;
        dot1qvlan_parms->callback_rw = CALLBACK_ACT_WRITE;
        snmp_call_callbacks(SNMP_CALLBACK_APPLICATION,
                                SNMPD_CALLBACK_BRIDGE,
                                dot1qvlan_parms);
        if (dot1qvlan_parms->rv == MIB_STATUS_ERROR) {
            AGENT_EXIT;
            return SNMP_ERR_COMMITFAILED;
        }
        break;
    }

    AGENT_EXIT;
    return SNMP_ERR_NOERROR;
}

int
write_dot1qPortIngressFiltering(int action,
                                u_char * var_val,
                                u_char var_val_type,
                                size_t var_val_len,
                                u_char * statP,
                                oid * name, size_t name_len)
{
    long            value;
    dot1qvlan_parms->index = name[DOT1Q_PORT_VLAN_ENTRY];

    switch (action) {
    case RESERVE1:
        if (var_val_type != ASN_INTEGER) {
            fprintf(stderr, "write to dot1qVlan not ASN_INTEGER\n");
            return SNMP_ERR_WRONGTYPE;
        }
        if (var_val_len > sizeof(long)) {
            fprintf(stderr, "write to dot1qVlan: bad length\n");
            return SNMP_ERR_WRONGLENGTH;
        }
        value = *(u_long *) var_val;
        if (value != MIB_STATUS_ENABLED && value != MIB_STATUS_DISABLED) {
            fprintf(stderr, "write to dot1qVlan: bad value\n");
            return SNMP_ERR_WRONGVALUE;
        }
        break;

    case RESERVE2:
        break;

    case FREE:
        break;

    case ACTION:
        break;

    case UNDO:
        break;

    case COMMIT:
        dot1qvlan_parms->value = * (long *) var_val;
        dot1qvlan_parms->group_obj = DOT1Q_PORTINGRESSFILTERING;
        dot1qvlan_parms->callback_rw = CALLBACK_ACT_WRITE;
        snmp_call_callbacks(SNMP_CALLBACK_APPLICATION,
                                SNMPD_CALLBACK_BRIDGE,
                                dot1qvlan_parms);
        if (dot1qvlan_parms->rv == MIB_STATUS_ERROR) {
            AGENT_EXIT;
            return SNMP_ERR_COMMITFAILED;
        }
        break;
    }

    AGENT_EXIT;
    return SNMP_ERR_NOERROR;
}

/*
 * No GVRP support currently
 */
int
write_dot1qPortGvrpStatus(int action,
                          u_char * var_val,
                          u_char var_val_type,
                          size_t var_val_len,
                          u_char * statP, oid * name, size_t name_len)
{
    long            value;
    dot1qvlan_parms->index = name[DOT1Q_PORT_VLAN_ENTRY];

    switch (action) {
    case RESERVE1:
        if (var_val_type != ASN_INTEGER) {
            fprintf(stderr, "write to dot1qVlan not ASN_INTEGER\n");
            return SNMP_ERR_WRONGTYPE;
        }
        if (var_val_len > sizeof(long)) {
            fprintf(stderr, "write to dot1qVlan: bad length\n");
            return SNMP_ERR_WRONGLENGTH;
        }
        value = *(u_long *) var_val;
        if (value != MIB_STATUS_ENABLED && value != MIB_STATUS_DISABLED) {
            fprintf(stderr, "write to dot1qVlan: bad value\n");
            return SNMP_ERR_WRONGVALUE;
        }
        break;

    case RESERVE2:
        break;

    case FREE:
        break;

    case ACTION:
        break;

    case UNDO:
        break;

    case COMMIT:
        dot1qvlan_parms->value = * (long *) var_val;
        dot1qvlan_parms->group_obj = DOT1Q_PORTGVRPSTATUS;
        dot1qvlan_parms->callback_rw = CALLBACK_ACT_WRITE;
        snmp_call_callbacks(SNMP_CALLBACK_APPLICATION,
                                SNMPD_CALLBACK_BRIDGE,
                                dot1qvlan_parms);
        if (dot1qvlan_parms->rv == MIB_STATUS_ERROR) {
            AGENT_EXIT;
            return SNMP_ERR_COMMITFAILED;
        }
        break;
    }

    AGENT_EXIT;
    return SNMP_ERR_NOERROR;
}

/*
 * VLAN constraints are not supported
 */
int
write_dot1qConstraintType(int action,
                          u_char * var_val,
                          u_char var_val_type,
                          size_t var_val_len,
                          u_char * statP, oid * name, size_t name_len)
{
    long            value;
    dot1qvlan_parms->index = name[DOT1Q_PORT_VLAN_ENTRY];

    switch (action) {
    case RESERVE1:
        if (var_val_type != ASN_INTEGER) {
            fprintf(stderr, "write to dot1qVlan not ASN_INTEGER\n");
            return SNMP_ERR_WRONGTYPE;
        }
        if (var_val_len > sizeof(long)) {
            fprintf(stderr, "write to dot1qVlan: bad length\n");
            return SNMP_ERR_WRONGLENGTH;
        }
        value = *(u_long *) var_val;
        if (value != 1 && value != 2) {
            fprintf(stderr, "write to dot1qVlan: bad value\n");
            return SNMP_ERR_WRONGVALUE;
        }
        break;

    case RESERVE2:
        /* Not supported */
        return SNMP_ERR_NOTWRITABLE;
        break;

    case FREE:
        break;

    case ACTION:
        break;

    case UNDO:
        break;

    case COMMIT:
        dot1qvlan_parms->value = * (long *) var_val;
        dot1qvlan_parms->group_obj = DOT1Q_CONSTRAINTTYPE;
        dot1qvlan_parms->callback_rw = CALLBACK_ACT_WRITE;
        snmp_call_callbacks(SNMP_CALLBACK_APPLICATION,
                                SNMPD_CALLBACK_BRIDGE,
                                dot1qvlan_parms);
        if (dot1qvlan_parms->rv == MIB_STATUS_ERROR) {
            AGENT_EXIT;
            return SNMP_ERR_COMMITFAILED;
        }
        break;
    }

    AGENT_EXIT;
    return SNMP_ERR_NOERROR;
}

/*
 * This is also not supported
 */
int
write_dot1qConstraintStatus(int action,
                            u_char * var_val,
                            u_char var_val_type,
                            size_t var_val_len,
                            u_char * statP, oid * name, size_t name_len)
{
    long            value;
    int             size;

    switch (action) {
    case RESERVE1:
        if (var_val_type != ASN_INTEGER) {
            fprintf(stderr, "write to dot1qVlan not ASN_INTEGER\n");
            return SNMP_ERR_WRONGTYPE;
        }
        if (var_val_len > sizeof(long)) {
            fprintf(stderr, "write to dot1qVlan: bad length\n");
            return SNMP_ERR_WRONGLENGTH;
        }
        break;

    case RESERVE2:
        /* Not supported */
        return SNMP_ERR_NOTWRITABLE;
        break;

    case FREE:
        /*
         * Release any resources that have been allocated 
         */
        break;

    case ACTION:
        /*
         * The variable has been stored in 'value' for you to use,
         * and you have just been asked to do something with it.
         * Note that anything done here must be reversable in the UNDO case
         */
        break;

    case UNDO:
        /*
         * Back out any changes made in the ACTION case 
         */
        break;

    case COMMIT:
        /*
         * Things are working well, so it's now safe to make the change
         * permanently.  Make sure that anything done here can't fail!
         */
        break;
    }
    return SNMP_ERR_NOERROR;
}

/*
 * This is also not supported
 */
int
write_dot1qConstraintSetDefault(int action,
                            u_char * var_val,
                            u_char var_val_type,
                            size_t var_val_len,
                            u_char * statP, oid * name, size_t name_len)
{
    long            value;
    int             size;

    switch (action) {
    case RESERVE1:
        if (var_val_type != ASN_INTEGER) {
            fprintf(stderr, "write to dot1qVlan not ASN_INTEGER\n");
            return SNMP_ERR_WRONGTYPE;
        }
        if (var_val_len > sizeof(long)) {
            fprintf(stderr, "write to dot1qVlan: bad length\n");
            return SNMP_ERR_WRONGLENGTH;
        }
        break;

    case RESERVE2:
        /* Not supported */
        return SNMP_ERR_NOTWRITABLE;
        break;

    case FREE:
        /*
         * Release any resources that have been allocated 
         */
        break;

    case ACTION:
        /*
         * The variable has been stored in 'value' for you to use,
         * and you have just been asked to do something with it.
         * Note that anything done here must be reversable in the UNDO case
         */
        break;

    case UNDO:
        /*
         * Back out any changes made in the ACTION case 
         */
        break;

    case COMMIT:
        /*
         * Things are working well, so it's now safe to make the change
         * permanently.  Make sure that anything done here can't fail!
         */
        break;
    }
    return SNMP_ERR_NOERROR;
}
int
write_dot1qConstraintTypeDefault(int action,
                            u_char * var_val,
                            u_char var_val_type,
                            size_t var_val_len,
                            u_char * statP, oid * name, size_t name_len)
{
    long            value;
    int             size;

    switch (action) {
    case RESERVE1:
        if (var_val_type != ASN_INTEGER) {
            fprintf(stderr, "write to dot1qVlan not ASN_INTEGER\n");
            return SNMP_ERR_WRONGTYPE;
        }
        if (var_val_len > sizeof(long)) {
            fprintf(stderr, "write to dot1qVlan: bad length\n");
            return SNMP_ERR_WRONGLENGTH;
        }
        break;

    case RESERVE2:
        /* Not supported */
        return SNMP_ERR_NOTWRITABLE;
        break;

    case FREE:
        /*
         * Release any resources that have been allocated 
         */
        break;

    case ACTION:
        /*
         * The variable has been stored in 'value' for you to use,
         * and you have just been asked to do something with it.
         * Note that anything done here must be reversable in the UNDO case
         */
        break;

    case UNDO:
        /*
         * Back out any changes made in the ACTION case 
         */
        break;

    case COMMIT:
        /*
         * Things are working well, so it's now safe to make the change
         * permanently.  Make sure that anything done here can't fail!
         */
        break;
    }
    return SNMP_ERR_NOERROR;
}

void free_dot1qVlan_createAndWaitMemory(void) 
{
    if (creation_row) {
        sal_free(creation_row);
        creation_row = NULL;
    }

    return;
}

