/*
 * $Id: egp.c,v 1.1.2.6 2006/10/18 06:05:58 bobyu Exp $
 */

/* TODO: Implement tables in this MIB once backedn support is present */
#include <config.h>


/* minimal include directives */
#include "mibincl.h"
#include "mibgroup/util_funcs.h"

#include <ucd-snmp/mibgroup/mibII/group_mib.h>
#include "mibgroup/mibII/egp.h"

/* Argument to be passed with callback */
static group_parameters_t *egpmib_parms = NULL;

/* Path to this MIB WRT the global MIB tree */
oid egp_variables_oid[] = { 1,3,6,1,2,1,8 };

/* Variables that this MIB supports */
struct variable4 egp_variables[] = {
/*  magic number        , variable type , ro/rw , callback fn  , L, oidsuffix */
#define EGPINMSGS		1
    {EGPINMSGS,  ASN_COUNTER,  RONLY ,  var_egp, 1,  { 1 }},
#define EGPINERRORS		2
    {EGPINERRORS,  ASN_COUNTER,  RONLY ,  var_egp, 1,  { 2 }},
#define EGPOUTMSGS		3
    {EGPOUTMSGS,  ASN_COUNTER,  RONLY ,  var_egp, 1,  { 3 }},
#define EGPOUTERRORS		4
    {EGPOUTERRORS,  ASN_COUNTER,  RONLY ,  var_egp, 1,  { 4 }},
#define EGPAS		5
    {EGPAS,  ASN_INTEGER,  RONLY ,  var_egp, 1,  { 6 }},

#define EGPNEIGHSTATE		6
    {EGPNEIGHSTATE,  ASN_INTEGER,  RONLY,   var_egpNeighTable, 3,  { 5, 1, 1 }},
#define EGPNEIGHADDR		7
    {EGPNEIGHADDR,  ASN_IPADDRESS,  RONLY,   var_egpNeighTable, 3,  { 5, 1, 2 }},
#define EGPNEIGHAS		8
    {EGPNEIGHAS,  ASN_INTEGER,  RONLY,   var_egpNeighTable, 3,  { 5, 1, 3 }},
#define EGPNEIGHINMSGS		9
    {EGPNEIGHINMSGS,  ASN_COUNTER,  RONLY,   var_egpNeighTable, 3,  { 5, 1, 4 }},
#define EGPNEIGHINERRS		10
    {EGPNEIGHINERRS,  ASN_COUNTER,  RONLY,   var_egpNeighTable, 3,  { 5, 1, 5 }},
#define EGPNEIGHOUTMSGS		11
    {EGPNEIGHOUTMSGS,  ASN_COUNTER,  RONLY,   var_egpNeighTable, 3,  { 5, 1, 6 }},
#define EGPNEIGHOUTERRS		12
    {EGPNEIGHOUTERRS,  ASN_COUNTER,  RONLY,   var_egpNeighTable, 3,  { 5, 1, 7 }},
#define EGPNEIGHINERRMSGS		13
    {EGPNEIGHINERRMSGS,  ASN_COUNTER,  RONLY,   var_egpNeighTable, 3,  { 5, 1, 8 }},
#define EGPNEIGHOUTERRMSGS		14
    {EGPNEIGHOUTERRMSGS,  ASN_COUNTER,  RONLY,   var_egpNeighTable, 3,  { 5, 1, 9 }},
#define EGPNEIGHSTATEUPS		15
    {EGPNEIGHSTATEUPS,  ASN_COUNTER,  RONLY,   var_egpNeighTable, 3,  { 5, 1, 10 }},
#define EGPNEIGHSTATEDOWNS		16
    {EGPNEIGHSTATEDOWNS,  ASN_COUNTER,  RONLY,   var_egpNeighTable, 3,  { 5, 1, 11 }},
#define EGPNEIGHINTERVALHELLO		17
    {EGPNEIGHINTERVALHELLO,  ASN_INTEGER,  RONLY,   var_egpNeighTable, 3,  { 5, 1, 12 }},
#define EGPNEIGHINTERVALPOLL		18
    {EGPNEIGHINTERVALPOLL,  ASN_INTEGER,  RONLY,   var_egpNeighTable, 3,  { 5, 1, 13 }},
#define EGPNEIGHMODE		19
    {EGPNEIGHMODE,  ASN_INTEGER,  RONLY,   var_egpNeighTable, 3,  { 5, 1, 14 }},
#define EGPNEIGHEVENTTRIGGER		20
    {EGPNEIGHEVENTTRIGGER,  ASN_INTEGER,  RWRITE,  var_egpNeighTable, 3,  { 5, 1, 15 }},
};

/** Initializes the egp module */
void
init_egp(void)
{
    AGENT_ENTER;

    /* Register MIB */
    egpmib_parms = (group_parameters_t *) \
                    sal_alloc(sizeof(group_parameters_t), "egpmib_parms");

    if (egpmib_parms != NULL) {
        /* register ourselves with the agent to handle our mib tree */
        REGISTER_MIB("egp", egp_variables, variable4,
                   egp_variables_oid);

        egpmib_parms->group = GROUP_EGP;
    }

    AGENT_EXIT;
}

/* This function contains all counter values */
unsigned char *
var_egp(struct variable *vp, 
                oid     *name, 
                size_t  *length, 
                int     exact, 
                size_t  *var_len, 
                WriteMethod **write_method)
{
    AGENT_ENTER;

    /* variables we may use later */
    static u_long ulong_ret;

    if (header_generic(vp,name,length,exact,var_len,write_method)
                                  == MATCH_FAILED ) {
        AGENT_EXIT;
        return NULL;
    }

    switch(vp->magic) {
        case EGPINMSGS:
            egpmib_parms->group_obj = EGP_INMSGS;
            break;
        case EGPINERRORS:
            egpmib_parms->group_obj = EGP_INERRORS;
            break;
        case EGPOUTMSGS:
            egpmib_parms->group_obj = EGP_OUTMSGS;
            break;
        case EGPOUTERRORS:
            egpmib_parms->group_obj = EGP_OUTERRORS;
            break;
        case EGPAS:
            egpmib_parms->group_obj = EGP_AS;
            break;
        default:
            AGENT_EXIT;
            return NULL;
    }

    egpmib_parms->callback_rw = CALLBACK_ACT_READ;
    egpmib_parms->group = GROUP_EGP; 
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                        SNMPD_CALLBACK_GROUPS,
                        egpmib_parms);

    ulong_ret = (unsigned long) egpmib_parms->value;
    AGENT_EXIT;
    return (unsigned char *) &ulong_ret;
}


/* EGP Neighbor table variables */
unsigned char *
var_egpNeighTable(struct variable *vp,
    	    oid     *name,
    	    size_t  *length,
    	    int     exact,
    	    size_t  *var_len,
    	    WriteMethod **write_method)
{
    /* variables we may use later */
    AGENT_ENTER;
    static long long_ret;
    static unsigned char string[SPRINT_MAX_LEN];

    /* For now, we return directly */
    AGENT_EXIT;
    return NULL;

    /* TODO: When egp support is provided, change value for TABLE_SIZE */
    if (header_simple_table(vp,name,length,exact,var_len,write_method, 1)
                                                == MATCH_FAILED ) {
        AGENT_EXIT;
        return NULL;
    }

    switch(vp->magic) {
        case EGPNEIGHSTATE:
            egpmib_parms->group_obj = EGP_NEIGHSTATE;
            break;
        case EGPNEIGHADDR:
            egpmib_parms->group_obj = EGP_NEIGHADDR;
            break;
        case EGPNEIGHAS:
            egpmib_parms->group_obj = EGP_NEIGHAS;
            break;
        case EGPNEIGHINMSGS:
            egpmib_parms->group_obj = EGP_NEIGHINMSGS;
            break;
        case EGPNEIGHINERRS:
            egpmib_parms->group_obj = EGP_NEIGHINERRS;
            break;
        case EGPNEIGHOUTMSGS:
            egpmib_parms->group_obj = EGP_NEIGHOUTMSGS;
            break;
        case EGPNEIGHOUTERRS:
            egpmib_parms->group_obj = EGP_NEIGHOUTERRS;
            break;
        case EGPNEIGHINERRMSGS:
            egpmib_parms->group_obj = EGP_NEIGHINERRMSGS;
            break;
        case EGPNEIGHOUTERRMSGS:
            egpmib_parms->group_obj = EGP_NEIGHOUTERRMSGS;
            break;
        case EGPNEIGHSTATEUPS:
            egpmib_parms->group_obj = EGP_NEIGHSTATEUPS;
            break;
        case EGPNEIGHSTATEDOWNS:
            egpmib_parms->group_obj = EGP_NEIGHSTATEDOWNS;
            break;
        case EGPNEIGHINTERVALHELLO:
            egpmib_parms->group_obj = EGP_NEIGHINTERVALHELLO;
            break;
        case EGPNEIGHINTERVALPOLL:
            egpmib_parms->group_obj = EGP_NEIGHINTERVALPOLL;
            break;
        case EGPNEIGHMODE:
            egpmib_parms->group_obj = EGP_NEIGHMODE;
            break;
        case EGPNEIGHEVENTTRIGGER:
#if 0 /* Not supported for now */
            *write_method = write_egpNeighEventTrigger;
#endif
            egpmib_parms->group_obj = EGP_NEIGHEVENTTRIGGER;
            break;
        default:
            ERROR_MSG("Undefined MIB reference in egp MIB");
    }

    /* Callback */
    egpmib_parms->callback_rw = CALLBACK_ACT_READ;
    egpmib_parms->group = GROUP_EGP; 
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                        SNMPD_CALLBACK_GROUPS,
                        egpmib_parms);
    switch (vp->magic) {
        case EGPNEIGHADDR:
            string[0] = string[1] = string[2] = string[3] = 0;
            *var_len = 4;
            return (unsigned char *) string;
    }

    long_ret = (long) egpmib_parms->value;
    AGENT_EXIT;
    return (unsigned char *) &long_ret;
}


int
write_egpNeighEventTrigger(int      action,
            u_char   *var_val,
            u_char   var_val_type,
            size_t   var_val_len,
            u_char   *statP,
            oid      *name,
            size_t   name_len)
{
    long value;
    int size;
    AGENT_ENTER;

    value = * (long *) var_val;
    switch ( action ) {
        case RESERVE1:
          if (var_val_type != ASN_INTEGER) {
              fprintf(stderr, "write to egp not ASN_INTEGER\n");
              AGENT_EXIT;
              return SNMP_ERR_WRONGTYPE;
          }
          if (var_val_len > sizeof(long)) {
              fprintf(stderr,"write to egp: bad length\n");
              AGENT_EXIT;
              return SNMP_ERR_WRONGLENGTH;
          }
          break;

        case RESERVE2:
          size  = var_val_len;
          break;

        case FREE:
             /* Release any resources that have been allocated */
          break;

        case ACTION:
          break;

        case UNDO:
             /* Back out any changes made in the ACTION case */
          break;

        case COMMIT:
          break;
    }

    AGENT_EXIT;
    return SNMP_ERR_NOERROR;
}
